
#include "arcball.h"
#include <math.h>

Vec3 MouseOnSphere(int mouse_x, int mouse_y, double w, double h, double ballRadius)
{
    Vec3 ballMouse;
    register double mag;
    ballMouse.x = (mouse_x - w) / ballRadius;
    ballMouse.y = (mouse_y - h) / ballRadius;
    mag = ballMouse.x*ballMouse.x + ballMouse.y*ballMouse.y;
    if (mag > 1.0) {
	register double scale = 1.0/sqrt(mag);
	ballMouse.x *= scale; ballMouse.y *= scale;
	ballMouse.z = 0.0;
    } else {
	ballMouse.z = sqrt(1 - mag);
    }
    return (ballMouse);
}

Quat Qt_FromBallPoints(Vec3 from, Vec3 to)
{
    Quat qu;
    qu.x = from.y*to.z - from.z*to.y;
    qu.y = from.z*to.x - from.x*to.z;
    qu.z = from.x*to.y - from.y*to.x;
    qu.w = from.x*to.x + from.y*to.y + from.z*to.z;
    return (qu);
}

Quat Qt_Mul(Quat qL, Quat qR)
{
    Quat qq;
    qq.w = qL.w*qR.w - qL.x*qR.x - qL.y*qR.y - qL.z*qR.z;
    qq.x = qL.w*qR.x + qL.x*qR.w + qL.y*qR.z - qL.z*qR.y;
    qq.y = qL.w*qR.y + qL.y*qR.w + qL.z*qR.x - qL.x*qR.z;
    qq.z = qL.w*qR.z + qL.z*qR.w + qL.x*qR.y - qL.y*qR.x;
    return (qq);
}

Mat4 *Qt_ToMatrix(Quat q, Mat4 out)
{
    double Nq = q.x*q.x + q.y*q.y + q.z*q.z + q.w*q.w;
    double s = (Nq > 0.0) ? (2.0 / Nq) : 0.0;
    double xs = q.x*s,	      ys = q.y*s,	  zs = q.z*s;
    double wx = q.w*xs,	      wy = q.w*ys,	  wz = q.w*zs;
    double xx = q.x*xs,	      xy = q.x*ys,	  xz = q.x*zs;
    double yy = q.y*ys,	      yz = q.y*zs,	  zz = q.z*zs;
    out[X][X] = 1.0 - (yy + zz); out[Y][X] = xy + wz; out[Z][X] = xz - wy;
    out[X][Y] = xy - wz; out[Y][Y] = 1.0 - (xx + zz); out[Z][Y] = yz + wx;
    out[X][Z] = xz + wy; out[Y][Z] = yz - wx; out[Z][Z] = 1.0 - (xx + yy);
    out[X][W] = out[Y][W] = out[Z][W] = out[W][X] = out[W][Y] = out[W][Z] = 0.0;
    out[W][W] = 1.0;
    return ((Mat4 *)&out);
}

Quat Qt_Conj(Quat q)
{
    Quat qq;
    qq.x = -q.x; qq.y = -q.y; qq.z = -q.z; qq.w = q.w;
    return (qq);
}


void transpose(Mat4 a)
{ 
	double t;					
	t = a[0][1]; a[0][1] = a[1][0]; a[1][0] = t;	
	t = a[0][2]; a[0][2] = a[2][0]; a[2][0] = t;	
	t = a[0][3]; a[0][3] = a[3][0]; a[3][0] = t;	
	t = a[1][2]; a[1][2] = a[2][1]; a[2][1] = t;	
	t = a[1][3]; a[1][3] = a[3][1]; a[3][1] = t;	
	t = a[2][3]; a[2][3] = a[3][2]; a[3][2] = t;		
}

void qt_from_matrix( Mat4 R, Quat* myq)
{
  double tr, s ,q[4];
  tr = R[0][0] + R[1][1] + R[2][2];
  if (tr >= 0) {
    s = sqrt(tr + 1) ;
    q[3] = s*.5 ;
    s = .5/s ;
    q[0] = (R[2][1] - R[1][2]) * s ;
    q[1] = (R[0][2] - R[2][0]) * s ;
    q[2] = (R[1][0] - R[0][1]) * s ;
  } else {
    int h = 0 ;
    if (R[1][1] > R[0][0]) h =1 ;
    if (R[2][2] > R[h][h]) h =2 ;
    switch(h) {
#define caseMacro(i,j,k) \
    case i: \
      s = sqrt((R[i][i] - (R[j][j]+R[k][k])) + 1) ; \
      q[i] = s*.5 ;\
      s = .5/s ;\
      q[j] = (R[i][j] + R[j][i])*s ;\
      q[k] = (R[i][k] + R[k][i])*s ;\
      q[3] = (R[k][j] - R[j][k])*s ;\
      break 
    caseMacro(0,1,2) ;
    caseMacro(1,2,0) ;
    caseMacro(2,0,1) ;
#undef caseMacro
    }
  }
  myq->x = q[0];
  myq->y = q[1];
  myq->z = q[2];
  myq->w = q[3];
}

Vec3 ConstrainToAxis(Vec3 loose, Vec3 axis)
{
    Vec3 onPlane;
    register float norm;
	onPlane.x = loose.x - axis.x * (axis.x*loose.x + axis.y*loose.y + axis.z*loose.z);
	onPlane.y = loose.y - axis.y * (axis.x*loose.x + axis.y*loose.y + axis.z*loose.z);
	onPlane.z = loose.z - axis.z * (axis.x*loose.x + axis.y*loose.y + axis.z*loose.z);
	norm = onPlane.x*onPlane.x + onPlane.y*onPlane.y + onPlane.z*onPlane.z;
	
    if (norm > 0.0) {
		if (onPlane.z < 0.0) 
		{
			onPlane.x *= -1;
			onPlane.y *= -1;
			onPlane.z *= -1;
		}
		double scale = 1/sqrt(norm);
		onPlane.x *= scale;
		onPlane.y *= scale;
		onPlane.z *= scale;
		return onPlane;
    } 
    if (axis.z == 1) {
		onPlane.x=1;onPlane.y=0;onPlane.z=0;
    } else {
		onPlane.x = -axis.y;
		onPlane.y = axis.x;
		onPlane.z = 0;
		norm = 1/sqrt(onPlane.x*onPlane.x + onPlane.y*onPlane.y + onPlane.z*onPlane.z);
		onPlane.x *= norm;onPlane.y *= norm;onPlane.x *= norm;
    }
    return (onPlane);
}
